@@ -29,8 +29,11 @@ struct sr_session_priv {
 	int hard_failure_count;
 	int submit_count;
 	sr_track_t *last_track;
+	int np_timer;
 };
 
+static void now_playing(sr_session_t *s, sr_track_t *t);
+
 sr_session_t *
 sr_session_new(const char *url,
 	       const char *client_id,
@@ -60,6 +63,9 @@ sr_session_free(sr_session_t *s)
 
 	priv = s->priv;
 
+	if (priv->np_timer)
+		g_source_remove(priv->np_timer);
+
 	soup_session_abort(priv->soup);
 	g_object_unref(priv->soup);
 	while (!g_queue_is_empty(priv->queue)) {
@@ -151,7 +157,7 @@ check_last(sr_session_t *s,
 		return;
 	playtime = timestamp - c->timestamp;
 	/* did the last track played long enough? */
-	if (playtime >= 240 || playtime >= c->length / 2)
+	if ((playtime >= 240 || playtime >= c->length / 2) && c->length > 30)
 		g_queue_push_tail(priv->queue, c);
 	else
 		sr_track_free(c);
@@ -167,18 +173,30 @@ sr_session_pause(sr_session_t *s)
 	g_mutex_unlock(priv->queue_mutex);
 }
 
+static gboolean
+do_now_playing(void *data)
+{
+	sr_session_t *s = data;
+	struct sr_session_priv *priv = s->priv;
+	priv->np_timer = 0;
+	now_playing(s, priv->last_track);
+	return FALSE;
+}
+
 void
 sr_session_add_track(sr_session_t *s,
 		     sr_track_t *t)
 {
 	struct sr_session_priv *priv = s->priv;
 
+	if (priv->np_timer)
+		g_source_remove(priv->np_timer);
+
+	priv->np_timer = g_timeout_add_seconds(3, do_now_playing, s);
+
 	g_mutex_lock(priv->queue_mutex);
 	check_last(s, t->timestamp);
-	if (t->length > 30)
-		priv->last_track = t;
-	else
-		sr_track_free(t);
+	priv->last_track = t;
 	g_mutex_unlock(priv->queue_mutex);
 }
 
@@ -593,3 +611,96 @@ sr_session_submit(sr_session_t *s)
 				   s);
 	g_string_free(data, false); /* soup gets ownership */
 }
+
+static void
+now_playing_cb(SoupSession *session,
+	       SoupMessage *message,
+	       gpointer user_data)
+{
+	sr_session_t *s = user_data;
+	const char *data, *end;
+
+	if (!SOUP_STATUS_IS_SUCCESSFUL(message->status_code))
+		/* now need to do anything drastic, right? */
+		return;
+
+	data = message->response_body->data;
+	end = strchr(data, '\n');
+	if (!end) /* really bad */
+		return;
+
+	if (strncmp(data, "BADSESSION", end - data) == 0)
+		invalidate_session(s);
+}
+
+#undef ADD_FIELD
+#define ADD_FIELD(id, fmt, field) \
+	do { \
+		if ((field)) \
+		g_string_append_printf(data, "&" id "=%" fmt, (field)); \
+		else \
+		g_string_append_printf(data, "&" id "="); \
+	} while(0);
+
+static void
+now_playing(sr_session_t *s,
+	    sr_track_t *t)
+{
+	struct sr_session_priv *priv = s->priv;
+	SoupMessage *message;
+	GString *data;
+	char *artist, *title;
+	char *album = NULL, *mbid = NULL;
+
+	/* haven't got the session yet? */
+	if (!priv->session_id)
+		return;
+
+	data = g_string_new(NULL);
+	g_string_append_printf(data, "s=%s", priv->session_id);
+
+	artist = soup_uri_encode(t->artist, EXTRA_URI_ENCODE_CHARS);
+	title = soup_uri_encode(t->title, EXTRA_URI_ENCODE_CHARS);
+	if (t->album)
+		album = soup_uri_encode(t->album, EXTRA_URI_ENCODE_CHARS);
+	if (t->mbid)
+		mbid = soup_uri_encode(t->mbid, EXTRA_URI_ENCODE_CHARS);
+
+	/* required fields */
+	g_string_append_printf(data, "&a=%s&t=%s", artist, title);
+
+	/* optional fields */
+	ADD_FIELD("b", "s", album);
+	ADD_FIELD("l", "i", t->length);
+	ADD_FIELD("n", "i", t->position);
+	ADD_FIELD("m", "s", mbid);
+
+	g_free(artist);
+	g_free(title);
+	g_free(album);
+	g_free(mbid);
+
+	message = soup_message_new("POST", priv->now_playing_url);
+	soup_message_set_request(message,
+				 "application/x-www-form-urlencoded",
+				 SOUP_MEMORY_TAKE,
+				 data->str,
+				 data->len);
+	soup_session_queue_message(priv->soup,
+				   message,
+				   now_playing_cb,
+				   s);
+	g_string_free(data, false); /* soup gets ownership */
+}
+
+void
+sr_session_set_proxy(sr_session_t *s, const char *url)
+{
+	struct sr_session_priv *priv = s->priv;
+	SoupURI *soup_uri;
+	if (url)
+		soup_uri = soup_uri_new(url);
+	else
+		soup_uri = NULL;
+	g_object_set(priv->soup, "proxy-uri", soup_uri, NULL);
+}
